"""
(C) Copyright 2012, 10gen

This is a label on a mattress. Do not modify this file!
"""

import threading, time, traceback, warnings, sets

warnings.simplefilter( 'ignore', DeprecationWarning )

getLogsAgentVersion = "1.4.1"

class GetLogsThread( threading.Thread ):
    """ When enabled pull log data from hosts. """

    def __init__( self, hostDef, mmsAgent):
        """ Initialize the object """
        threading.Thread.__init__( self )
        self.hostDef = hostDef
        self.mmsAgent = mmsAgent
        self.logger = mmsAgent.logger
        self.logConn = None
        self.passes = 0
        self.lastLogEntries = sets.Set()

    def run( self ):
        """ Pull the data from the various hosts. """

        hostKey = self.hostDef['hostKey']

        self.logger.info( 'starting log data collection: ' + hostKey )

        sleepTime = self.mmsAgent.logCollectionInterval

        if ( sleepTime < 1 ):
            sleepTime = 5

        while not self.mmsAgent.done and self.hostDef['threadRunning'] and self.hostDef['getLogs']:
            try:
                self.passes = self.passes + 1

                time.sleep( sleepTime )

                if not self.mmsAgent.haveHostDef( hostKey ):
                    continue

                # Close the connection every so often.
                if self.passes % 120 == 0:
                    self.logConn = self.mmsAgent.closeDbConnection( self.hostDef, self.logConn )

                if not self.logConn:
                    self.logConn = self.mmsAgent.getDbConnection( self.hostDef )

                # Verify the connection.
                if not self.mmsAgent.isValidMonitorConn( self.hostDef, self.logConn ):
                    self.logConn = None
                    continue

                if not self.mmsAgent.hasCommand( 'getLog', self.hostDef, self.logConn ):
                    self.logger.error( 'This version of MongoDB does not support the getLog command' )
                    time.sleep( 60 )
                    continue

                logs = self._collectLogs( )

                # Make sure we ended up with the same connection.
                if not self.mmsAgent.isValidMonitorConn( self.hostDef, self.logConn ):
                    self.logConn = None
                    continue

                self.mmsAgent.setHostState( hostKey, 'logs', logs )

            except Exception, e:

                self.logConn = self.mmsAgent.closeDbConnection( self.hostDef, self.logConn )

                self.logger.error( 'Problem collecting log data from: ' + hostKey + " - exception: " + traceback.format_exc( e ) )

        self.logger.info( 'stopping log data collection: ' + hostKey )

        if self.logConn is not None:
            self.logConn.disconnect()
            self.logConn = None

    def _collectLogs( self ):
        """ Make the call to mongo host and pull the log data """
        root = { }

        logEntries = []

        entries = self.logConn.admin.command( { 'getLog' : 'global' } )

        if not entries or not entries.get('log'):
            self.lastLogEntries.clear()
            return root

        for logEntry in entries['log']:
            if logEntry in self.lastLogEntries:
                continue

            logEntries.append( logEntry )

        if len ( logEntries ) == 0:
            return root

        self.lastLogEntries.clear()

        for logEntry in entries['log']:
            self.lastLogEntries.add( logEntry )

        root['entries'] = logEntries

        return root

